/* $EPIC:$ */
/*
 * ircaux.h: header file for ircaux.c 
 *
 * Copyright (c) 1990 Michael Sandroff.
 * Copyright (c) 1991, 1992 Troy Rollo.
 * Copyright (c) 1992-1996 Matthew Green.
 * Copyright  1993, 1997 Jeremy Nelson.
 * Copyright  1994 Jake Khuon.
 * Copyright  1995, 2000 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#ifndef __ircaux_h__
#define __ircaux_h__

#define SAFE(x) (((x) && *(x)) ? (x) : empty_string)

typedef int 	comp_len_func 		(char *, char *, int);
typedef int 	comp_func 		(char *, char *);

#define new_malloc(x) 	really_new_malloc ((x), __FILE__, __LINE__)
#define new_free(x)   	really_new_free	  ((void **)(x), __FILE__, __LINE__)
#define MUST_BE_MALLOCED(x, y) \
		fatal_malloc_check ((void *)(x), (y), __FILE__, __LINE__)
#define RESIZE(x, y, z) new_realloc	  ((void **)& (x), sizeof(y) * (z))
#define LOCAL_COPY(y) 	strcpy		  (alloca(strlen((y)) + 1), y)

	void	fatal_malloc_check	(void *, Char *, char *, int);
	void *	really_new_malloc 	(size_t, char *, int);
	void *	really_new_free 	(void **, char *, int);
	int	debug_new_free		(void **, char *, int);
	void *	new_realloc 		(void **, size_t);

	char *	check_nickname 		(char *, int);
	char *	next_arg 		(char *, char **);
	char *	new_next_arg 		(char *, char **);
	char *	new_new_next_arg 	(char *, char **, char *);
	char *	s_next_arg		(char **);
	char *	last_arg 		(char **);
	char *	expand_twiddle 		(char *);
	char *	upper 			(char *);
	char *	lower 			(char *);
	char *	sindex			(char *, char *);
	char *	rsindex 		(char *, char *, char *, int);
	char *	path_search 		(char *, char *);
	char *	double_quote 		(Char *, Char *, char *);
	char *	malloc_strcpy 		(char **, Char *);
	char *	malloc_strcat 		(char **, Char *);
	char *	malloc_str2cpy		(char **, Char *, Char *);
	char *	m_s3cat_s 		(char **, Char *, Char *);
	char *	m_s3cat 		(char **, Char *, Char *);
	char *	m_3cat 			(char **, Char *, Char *);
	char *	m_e3cat 		(char **, Char *, Char *);
	char *	m_2dup 			(Char *, Char *);
	char *	m_3dup 			(Char *, Char *, Char *);
	char *	m_opendup 		(Char *, ...) __A(1);
	char *	m_strdup 		(Char *);
	void	wait_new_free 		(char **);
	char *	malloc_sprintf 		(char **, Char *, ...) __A(2);
	char *	m_sprintf 		(Char *, ...) __A(1);
	int	is_number 		(Char *);
	int	is_real_number		(Char *);
	char *	my_ctime 		(time_t);
extern	Uchar 	stricmp_table[];
	int	my_stricmp 		(Uchar *, Uchar *);
	int	my_strnicmp 		(Uchar *, Uchar *, size_t);
	int	scanstr 		(char *, char *);
	void	really_free 		(int);
	char *	chop 			(char *, size_t);
	char *	strmcpy 		(char *, Char *, int);
	char *	strmcat 		(char *, Char *, int);
	char *	strmcat_ue 		(char *, Char *, int);
	char *	m_strcat_ues 		(char **, Char *, int);
	char *	strmopencat		(char *, int, ...);
	char *	stristr 		(Char *, Char *);
	char *	rstristr 		(Char *, Char *);
	FILE *	uzfopen 		(char **, char *, int);
	int	end_strcmp 		(Char *, Char *, int);
	void	panic 			(char *, ...) __A(1) __N;
	int	fw_strcmp 		(comp_len_func *, char *, char *);
	int	lw_strcmp 		(comp_func *, char *, char *);
	int	open_to 		(char *, int, int);
	Timeval	get_time 		(Timeval *);
	double 	time_diff 		(Timeval, Timeval);
	char *	plural 			(int);
	int	time_to_next_minute 	(void);
	char *	remove_trailing_spaces 	(char *);
	char *	ltoa 			(long);
	char *	ftoa			(double);
	char *	strformat 		(char *, Char *, int, int);
	char *	chop_word 		(char *);
	int	splitw 			(char *, char ***);
	char *	unsplitw 		(char ***, int);
	int	check_val 		(const char *);
	char *	strext			(const char *,const char *);
	char *	strextend 		(char *, char, int);
	char *	pullstr 		(char *, char *);
	int 	empty 			(Char *);
	char *	safe_new_next_arg 	(char *, char **);
	char *	MatchingBracket 	(char *, char, char);
	int	word_count 		(Char *);
	int	parse_number 		(char **);
	char *	remove_brackets 	(Char *, Char *, int *);
	long	my_atol 		(Char *);
	char *	m_dupchar 		(int);
	char *	strmccat		(char *, char, int);
	off_t	file_size		(char *);
	int	is_root			(Char *, Char *, int);
	size_t	streq			(Char *, Char *);
	char *	m_strndup		(Char *, size_t);
	char *	prntdump		(Char *, size_t);
	char *	ov_strcpy		(char *, Char *);
	size_t	ccspan			(Char *, int);
	int	last_char		(Char *);
#ifndef HAVE_VSNPRINTF
	int	vsnprintf 		(char *, size_t, Char *, va_list);
#endif
#ifndef HAVE_SNPRINTF
	int	snprintf 		(char *, size_t, Char *, ...) __A(3);
#endif
	char *	next_in_comma_list	(char *, char **);
	char *	get_userhost		(void);
	int	charcount		(Char *, char);
	void	beep_em			(int);
	void	strip_control		(Char *, char *);
	Char *	strfill		(char, int);
	char *	encode			(Char *, int);
	char *	decode			(Char *);
	char *	chomp			(char *);
	int 	opento			(Char *, int, off_t);
	int	figure_out_address	(char *, char **, char **, 
					 char **, char **, int *);
	int	count_char		(Uchar *, Uchar);
	char *	strnrchr		(char *, char, int);
	void	mask_digits		(char **);
	char *	strpcat			(char *, Char *, ...) __A(2);
	char *	strmpcat		(char *, size_t, Char *, ...) __A(3);
	u_char *strcpy_nocolorcodes	(u_char *, Uchar *);
	u_long	random_number		(u_long);
	char *	urlencode		(Char *);
	char *	urldecode		(char *);
	char *	next_line		(char *, char **);
	char *	canonicalize_number	(char *);
	Char *	find_forward_quote	(Char *, Char *);
	Char *	find_backward_quote	(Char *, Char *);
	Char *	my_strerror		(int);

/* From words.c */
#define SOS 			-32767
#define EOS 		 	 32767
	char *	search 			(char *, char **, char *, int);
	Char *	real_move_to_abs_word 	(Char *, Char **, int, int);
	char *	real_extract 		(char *, int, int, int);
	char *	real_extract2 		(Char *, int, int, int);
#define move_to_abs_word(a, b, c)	real_move_to_abs_word(a, b, c, 0)
#define extract(a, b, c)		real_extract(a, b, c, 0)
#define extract2(a, b, c)		real_extract2(a, b, c, 0)
#define extractw(a, b, c)		real_extract(a, b, c, 1)
#define extractw2(a, b, c)		real_extract2(a, b, c, 1)

extern	unsigned char	isspace_table[256];
#define my_isspace(x)	isspace_table[(unsigned)(unsigned char)(x)]
#define my_isdigit(x) \
	(isdigit(*x) || ((*x == '-' || *x == '+') && isdigit(x[1])))
#define my_bzero(x) memset ( &(x) , 0 , sizeof(x) );

/* Used for the inbound mangling stuff */
#define MANGLE_ESCAPES		1 << 0
#define MANGLE_ANSI_CODES	1 << 1
#define STRIP_COLOR		1 << 2
#define STRIP_REVERSE		1 << 3
#define STRIP_UNDERLINE		1 << 4
#define STRIP_BOLD		1 << 5
#define STRIP_BLINK		1 << 6
#define STRIP_ROM_CHAR		1 << 7
#define STRIP_ND_SPACE		1 << 8
#define STRIP_ALT_CHAR		1 << 9
#define STRIP_ALL_OFF		1 << 10

extern	int	outbound_line_mangler;
extern	int	inbound_line_mangler;
	size_t		mangle_line		(char *, int, size_t);

/* Used from compat.c */
#ifndef HAVE_TPARM
	char *		tparm (const char *, ...);
#endif
#ifndef HAVE_STRTOUL
	u_long		strtoul (const char *, char **, int);
#endif
	char *		bsd_getenv (const char *);
	int		bsd_putenv (const char *);
	int		bsd_setenv (const char *, const char *, int);
	void		bsd_unsetenv (const char *);
#ifndef HAVE_INET_ATON
	int		inet_aton (const char *, struct in_addr *);
#endif
#ifndef HAVE_STRLCPY
	size_t		strlcpy (char *, const char *, size_t);
#endif
#ifndef HAVE_STRLCAT
	size_t        	strlcat (char *, const char *, size_t);
#endif
#ifndef HAVE_VSNPRINTF
	int   		vsnprintf (char *, size_t, const char *, va_list);
#endif
#ifndef HAVE_SNPRINTF
	int   		snprintf (char *, size_t, const char *, ...);
#endif
#ifndef HAVE_SETSID
	int   		setsid (void);
#endif
#ifndef STPCPY_DECLARED
	char *		stpcpy (char *, const char *);
#endif
	void		bsd_arc4random_stir (void);
	void		bsd_arc4random_addrandom (u_char *, int);
	u_32int_t	bsd_arc4random (void);

#endif /* _IRCAUX_H_ */
