/* $EPIC:$ */
/*
 * server.c:  Things dealing with that wacky program we call ircd.
 *
 * Copyright (c) 1990 Michael Sandroff.
 * Copyright (c) 1991, 1992 Troy Rollo.
 * Copyright (c) 1992-1996 Matthew Green.
 * Copyright  1993-2000 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#define NEED_SERVER_LIST
#include "irc.h"
#include "parse.h"
#include "server.h"
#include "ircaux.h"
#include "lastlog.h"
#include "exec.h"
#include "window.h"
#include "output.h"
#include "names.h"
#include "hook.h"
#include "notify.h"
#include "screen.h"
#include "status.h"
#include "vars.h"
#include "newio2.h"
#include "alias.h"
#include "commands.h"
#include "network.h"

#define MAXIMUM_SERVERS		128

/*
 * Note to future maintainers -- we do a bit of chicanery here.  The 'flags'
 * member in the server structure is assumed to be at least N bits wide,
 * where N is the number of possible user modes.  Since the server stores
 * our user modes in a similar fashion, this shouldnt ever be "broken",
 * and if it is, we'll just change to do it however the server does.
 * The easiest way to handle that would be just to use an fd_set.
 *
 * 'umodes' here is a bogus default.  When we recieve the 004 numeric, it
 * overrides 'umodes'.
 *
 * The 'o' user mode is a special case, in that we want to know when its on.
 * This is mirrored in the "operator" member, for historical reasons.  This is
 * a kludge and should be changed.
 */
const 	char *	umodes = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
const	char *	cmodes = "aciklmnprstzR";

static	char *	do_umode (int du_index);
	void 	reset_nickname (int);

	int	never_connected = 1;		/* true until first connection
						 * is made */
	int	connected_to_server = 0;	/* true when connection is
						 * confirmed */

static	int	block_server_delete = 0;
static	char	lame_wait_nick[] = "***LW***";
static	char	wait_nick[] = "***W***";

	Server	*server_list[MAXIMUM_SERVERS];
	int	number_of_servers = 0;
	int	max_servref = 0;

	int	primary_server = -1;
	int	from_server = -1;
	int	parsing_server_index = -1;
	int	last_server = -1;


/* server_list_size: returns the number of servers in the server list */
int 	server_list_size (void)
{
	return max_servref;
}

/************************ SERVERLIST STUFF ***************************/

/*
 * add_to_server_list: adds the given server to the server_list.  If the
 * server is already in the server list it is not re-added... however, if the
 * overwrite flag is true, the port and passwords are updated to the values
 * passes.  If the server is not on the list, it is added to the end. In
 * either case, the server is made the current server. 
 */
void 	add_to_server_list (Char *server, int port, Char *password, Char *nick, int overwrite)
{
	Server *s;

	if ((from_server = find_in_server_list(server, port)) == -1)
	{
		from_server = max_servref++;
		number_of_servers++;
		server_list[from_server] = (Server *)new_malloc(sizeof(Server));

		s = get_server(from_server);
		s->name = m_strdup(server);
		s->itsname = (char *) 0;
		s->password = (char *) 0;
		s->away = (char *) 0;
		s->version_string = (char *) 0;
		s->server2_8 = 0;
		s->oper = 0;
		s->des = -1;
		s->version = 0;
		s->flags = 0;
		s->flags2 = 0;
		s->nickname = (char *) 0;
		s->s_nickname = (char *) 0;
		s->d_nickname = (char *) 0;
		s->userhost = (char *) 0;
		s->connected = 0;
		s->connection_in_progress = 0;
		s->eof = 0;
		s->motd = 1;
		s->port = port;
		s->who_queue = NULL;
		s->ison_queue = NULL;
		s->userhost_queue = NULL;
		ADDRESS(s->local_addr) = 0;
		ADDRESS(s->uh_addr) = 0;
		s->umodes = NULL;
		s->redirect = NULL;
		s->cookie = NULL;
		s->closing = 0;
		s->nickname_pending = 0;
		s->fudge_factor = 0;
		s->registration_pending = 0;
		s->resetting_nickname = 0;
		s->doing_privmsg = 0;
		s->doing_notice = 0;
		s->in_ctcp_flag = 0;
		s->waiting_in = 0;
		s->waiting_out = 0;
		s->start_wait_list = NULL;
		s->end_wait_list = NULL;
		s->invite_channel = NULL;
		s->last_notify_nick = NULL;
		s->joined_nick = NULL;
		s->public_nick = NULL;
		s->recv_nick = NULL;
		s->sent_nick = NULL;
		s->sent_body = NULL;
		s->channels = init_channels();

		if (password && *password)
			malloc_strcpy(&s->password, password);
		if (nick && *nick)
			malloc_strcpy(&s->d_nickname, nick);
		else if (!s->d_nickname)
			malloc_strcpy(&s->d_nickname, nickname);
		malloc_strcpy(&s->umodes, umodes);

		make_notify_list(from_server);
		do_umode(from_server);
	}
	else
	{
		s = get_server(from_server);

		if (overwrite)
		{
			s->port = port;
			if (password || !s->password)
			{
				if (password && *password)
					malloc_strcpy(&s->password, password);
				else
					new_free(&s->password);
			}
			if (nick || !s->d_nickname)
			{
				if (nick && *nick)
					malloc_strcpy(&s->d_nickname, nick);
				else
					new_free(&s->d_nickname);
			}
		}

		if (strlen(server) > strlen(s->name))
			malloc_strcpy(&s->name, server);
	}
}

#define OUT_OF_BOUNDS(i)  if (i < 0 || i >= max_servref || !SERVER(i))

static 	void 	remove_from_server_list (int i)
{
	Server  *s;

	OUT_OF_BOUNDS(i)
		return;

	s = get_server(i);

	if (block_server_delete)
	{
		say("You can't delete server [%d] here!", i);
		return;
	}

	if (number_of_servers == 1)
	{
		say("You can't delete the last server!");
		return;
	}

	if (s->windows)
	{
		say("You can't delete a server with windows!");
		return;
	}

	if (i == 0)
	{
		say("You can't delete server refnum 0! (sorry)");
		return;
	}

	say("Deleting server [%d]", i);
	clean_server_queues(i);
	new_free(&s->name);
	new_free(&s->itsname);
	new_free(&s->password);
	new_free(&s->away);
	new_free(&s->version_string);
	new_free(&s->nickname);
	new_free(&s->s_nickname);
	new_free(&s->d_nickname);
	new_free(&s->userhost);
	new_free(&s->cookie);
	channel_server_destroy(i);
	exec_server_delete(i);
	new_free((char **)&s);
	server_list[i] = NULL;

	if (i == primary_server)
		primary_server = 0;
	if (i == from_server)
		from_server = 0;
	if (i == max_servref)
		max_servref--;
	number_of_servers--;
}


Server *	traverse_all_servers (int *counter)
{
	do 
		(*counter)++;
	while (*counter < max_servref && !SERVER(*counter));

	if (*counter == max_servref)
		return NULL;

	return SERVER(*counter);
}

/*
 * Given a hostname and a port number, this function returns the server_list
 * index for that server.  If the specified server is not in the server_list,
 * then -1 is returned.  This function makes an attempt to do smart name
 * completion so that if you ask for "irc", it returns the first server that
 * has the leading prefix of "irc" in it.
 */
Servref	find_in_server_list (const char *server, int port)
{
	Servref		i = -1;
	Servref		first_idx = -1;
	Server *	s;
	Char *		name;
	Char *		itsname;

	while ((s = traverse_all_servers(&i)))
	{
		/* Check the port first.  This is a cheap, easy check */
		if (port && s->port && port != s->port && port != -1)
			continue;

		name = get_server_name(i);
		itsname = get_server_itsname(i);

#define MATCH_WITH_COMPLETION(n1, n2)                         \
{                                                             \
      /* Length of the user's input */                        \
      size_t l1 = strlen(n1);                                 \
                                                              \
      /* Length of the server's real name */                  \
      size_t l2 = strlen(n2);                                 \
                                                              \
      /*                                                      \
       * Compare what the user wants to what this server      \
       * name is.  If the server's name is shorter than what  \
       * the user specified, then don't bother doing the      \
       * compare.  If the two strings match exactly, then     \
       * this is the server we're looking for.                \
       */                                                     \
      if (l2 >= l1 && !my_strnicmp(n1, n2, l1))               \
      {                                                       \
              if (l2 == l1)                                   \
                      return i;                               \
              if (first_idx == -1)                            \
                      first_idx = i;                          \
      }                                                       \
}

		MATCH_WITH_COMPLETION(server, name)
		if (!itsname)
			continue;
		MATCH_WITH_COMPLETION(server, itsname)
	}

	return first_idx;		/* -1 if server was not found */
}

/*
 * Given a string that (in all likelihood) contains a server description
 * of the form:
 *
 *		refnum	 (where refnum is an integer for an existing server)
 * or
 *		hostname:port:password:nickname
 * or
 *		hostname port password nickname
 *
 * This extracts the salient information from the string and returns the
 * server_list index for that server.  If the information describes a server
 * that is not in the server_list, that information is *automatically added*
 * to the server_list and a new index is returned.
 *
 * This function always succeeds.
 */
int 	find_server_refnum (char *server, char **rest)
{
	Servref 	refnum;
	int		port = irc_port;
	char *		cport = NULL;
	char *		password = NULL;
	char *		nick = NULL;

	/*
	 * First of all, check for an existing server refnum
	 */
	if ((refnum = parse_server_index(server)) != -1)
		return refnum;

	/*
	 * Next check to see if its a "server:port:password:nick"
	 */
	else if (strchr(server, ':'))
		parse_server_info(server, &cport, &password, &nick);

	/*
	 * Next check to see if its "server port password nick"
	 */
	else if (rest && *rest)
	{
		cport = next_arg(*rest, rest);
		password = next_arg(*rest, rest);
		nick = next_arg(*rest, rest);
	}

	if (cport && *cport)
		port = my_atol(cport);

	/*
	 * Add to the server list (this will update the port
	 * and password fields).
	 */
	add_to_server_list(server, port, password, nick, 1);
	return from_server;
}


/*
 * get_servref_by_desc: given a string that may contain a server refnum or
 * the name of a server, return the server index that corresponds to that
 * server description.  Returns -1 if the description does not fit any
 * valid server.
 */
int	get_servref_by_desc (const char *str)
{
	Servref	i;
	Char *	s;

	if (is_number(str))
	{
		if ((i = parse_server_index(str)) != -1)
			return i;
		return -1;
	}

	for (i = 0; i < number_of_servers; i++)
	{
		if ((s = get_server_itsname(i)) && !my_stricmp(str, s))
			return i;
		else if (!my_stricmp(str, get_server_name(i)))
			return i;
	}
	return -1;
}

/*
 * parse_server_index:  given a string, this checks if it's a number, and if
 * so checks it validity as a server index.  Otherwise -1 is returned 
 */
Servref	parse_server_index (const char *str)
{
	int	i;

	if (!is_number(str))
		return -1;
	else
	{
		i = atoi(str);
		OUT_OF_BOUNDS(i)
			return -1;
		return i;
	}
}

/*
 * parse_server_info:  This parses a single string of the form
 * "server:portnum:password:nickname".  It the points port to the portnum
 * portion and password to the password portion.  This chews up the original
 * string, so * upon return, name will only point the the name.  If portnum
 * or password are missing or empty,  their respective returned value will
 * point to null. 
 */
void	parse_server_info (char *name, char **port, char **password, char **nick)
{
	char *ptr;

	*port = *password = *nick = NULL;

	do
	{
		if (!(ptr = strchr(name, ':')))
			break;
		*ptr++ = 0;
		if (!*ptr)
			break;
		*port = ptr;

		if (!(ptr = strchr(ptr, ':')))
			break;
		*ptr++ = 0;
		if (!*ptr)
			break;
		*password = ptr;

		if (!(ptr = strchr(ptr, ':')))
			break;
		*ptr++ = 0;
		if (!*ptr)
			break;
		*nick = ptr;

		/* Ignore any additional, future fields */
		if ((ptr = strchr(ptr, ':')) != NULL)
			*ptr++ = 0;
	}
	while (0);
}

/*
 * build_server_list: given a whitespace separated list of server names this
 * builds a list of those servers using add_to_server_list().  Since
 * add_to_server_list() is used to added each server specification, this can
 * be called many many times to add more servers to the server list.  Each
 * element in the server list case have one of the following forms: 
 *
 * servername 
 *
 * servername:port 
 *
 * servername:port:password 
 *
 * servername::password 
 *
 * Note also that this routine mucks around with the server string passed to it,
 * so make sure this is ok 
 */
void	build_server_list (char *servers)
{
	char	*host,
		*rest,
		*password = (char *) 0,
		*port = (char *) 0,
		*nick = (char *) 0;
	int	port_num;

	if (!servers)
		return;

	while (servers)
	{
		if ((rest = strchr(servers, '\n')))
			*rest++ = 0;

		while ((host = next_arg(servers, &servers)))
		{
			parse_server_info(host, &port, &password, &nick);
                        if (port && *port && (port_num = my_atol(port)))
				;
			else
				port_num = irc_port;

			add_to_server_list(host, port_num, password, nick, 0);
		}
		servers = rest;
	}
}

/*
 * read_server_file: reads hostname:portnum:password server information from
 * a file and adds this stuff to the server list.  See build_server_list()/ 
 */
int 	read_server_file (void)
{
	FILE 	*fp;
	char	file_path[MAXPATHLEN + 1];
	char	buffer[BIG_BUFFER_SIZE + 1];
	char	*expanded;

	if (getenv("IRC_SERVERS_FILE"))
		strmcpy(file_path, getenv("IRC_SERVERS_FILE"), MAXPATHLEN);
	else
	{
#ifdef SERVERS_FILE
		*file_path = 0;
		if (SERVERS_FILE[0] != '/' && SERVERS_FILE[0] != '~')
			strmcpy(file_path, irc_lib, MAXPATHLEN);
		strmcat(file_path, SERVERS_FILE, MAXPATHLEN);
#else
		return 1;
#endif
	}

	expanded = expand_twiddle(file_path);
	if ((fp = fopen(expanded, "r")))
	{
		while (fgets(buffer, BIG_BUFFER_SIZE, fp))
		{
			chop(buffer, 1);
			build_server_list(buffer);
		}
		fclose(fp);
		new_free(&expanded);
		return (0);
	}

	new_free(&expanded);
	return 1;
}

/* display_server_list: just guess what this does */
void 	display_server_list (void)
{
	int		i = -1;
	Server *	s;

	if (from_server != -1)
		say("Current server: %s %d",
				get_server_name(from_server),
				get_server_port(from_server));
	else
		say("Current server: <None>");

	if (primary_server != -1)
		say("Primary server: %s %d",
			get_server_name(primary_server),
			get_server_port(primary_server));
	else
		say("Primary server: <None>");


	say("Server list:");
	while ((s = traverse_all_servers(&i)))
	{
		if (!s->nickname)
			say("\t%d) %s %d", i, 
				get_server_name(i), 
				get_server_port(i));
		else
		{
			if (is_server_open(i))
				say("\t%d) %s %d (was %s)", i,
					get_server_name(i), 
					get_server_port(i),
					get_server_nickname(i));
			else
				say("\t%d) %s %d (%s)", i,
					get_server_name(i), 
					get_server_port(i),
					get_server_nickname(i));
		}
	}
}

char *	create_server_list (void)
{
	int	i = -1;
	char *	buffer = NULL;
	Server *	s;

	while ((s = traverse_all_servers(&i)))
	{
		if (s->des != -1)
		{
			if (s->itsname)
				m_s3cat(&buffer, space, s->itsname);
			else
				yell("Warning: server_list[%d]->itsname is "
					"null and it shouldnt be", i);
		}
	}

	return buffer ? buffer : m_strdup(empty_string);
}

/**************************** BUILT IN COMMANDS **************************/
/*
 * server: the /SERVER command. Read the SERVER help page about 
 */
BUILT_IN_COMMAND(servercmd)
{
	char	*server = NULL;
	int	i;

	if ((server = next_arg(args, &args)) == NULL)
	{
		display_server_list();
		return;
	}

	/*
	 * Delete an existing server
	 */
	if (strlen(server) > 1 && 
		!my_strnicmp(server, "-DELETE", strlen(server)))
	{
		if (block_server_delete > 0)
		{
			say("You may not do /SERVER -DELETE here, sorry.");
			return;
		}

		if ((server = next_arg(args, &args)) == NULL)
		{
			say("Need server number for -DELETE");
			return;
		}

		if ((i = parse_server_index(server)) == -1 &&
		    (i = find_in_server_list(server, 0)) == -1)
		{
			say("No such server in list");
			return;
		}

		if (is_server_open(i))
		{
			say("You may not delete a server that is open");
			return;
		}

		remove_from_server_list(i);
	}

	/*
	 * Add a server, but dont connect
	 */
	else if (strlen(server) > 1 && 
			!my_strnicmp(server, "-ADD", strlen(server)))
	{
		if (!(server = new_next_arg(args, &args)))
		{
			say("Need server info for -ADD");
			return;
		}

		find_server_refnum(server, &args);
	}

	/*
	 * The difference between /server +foo.bar.com and
	 * /window server foo.bar.com is now moot.
	 */
	else if (*server == '+')
	{
		/* /SERVER +foo.bar.com is an alias for /window server */
		if (*++server)
			window_server(current_window, &server);

		/* /SERVER + means go to the next server */
		else
			get_connected(from_server + 1, from_server);

		window_check_servers();
	}

	/*
	 * You can only detach a server using its refnum here.
	 */
	else if (*server == '-')
	{
		if (*++server)
		{
			i = find_server_refnum(server, &args);
			if (i == primary_server)
			{
			    say("You can't close your primary server!");
			    return;
			}
			close_server(i, "closing server");
		}
		else
			get_connected(from_server - 1, from_server);

		window_check_servers();
	}

	/*
	 * Just a naked /server with no flags
	 */
	else
	{
		int j = from_server;

		i = find_server_refnum(server, &args);
		connect_to_new_server(i, j, 0);
		window_check_servers();
	}
}

BUILT_IN_COMMAND(disconnectcmd)
{
	char	*server;
	char	*message;
	int	i;

	if (!(server = next_arg(args, &args)))
		i = get_window_server(0);
	else
	{
		if ((i = parse_server_index(server)) == -1)
		{
			say("No such server!");
			return;
		}
	}

	if (i >= 0 && i < number_of_servers)
	{
		if (!args || !*args)
			message = "Disconnecting";
		else
			message = args;

		say("Disconnecting from server %s", get_server_itsname(i));
		close_server(i, message);
		window_check_servers();
	}

	if (!connected_to_server)
                if (do_hook(DISCONNECT_LIST, "Disconnected by user request"))
			say("You are not connected to a server, use /SERVER to connect.");
} 


/********************* INPUT/OUTPUT STUFF ***********************/
int	server_error (int fd)
{
	int	sopen;
	int	i = -1;
	Server *s;

	while ((s = traverse_all_servers(&i)))
		if (s->des == fd)
			break;

	last_server = from_server = i;

	sopen = is_server_connected(i);
	close_server(i, empty_string);
	say("Connection closed from %s: %s", 
		get_server_name(i), (io_errno == -1) ? 
			     "Remote end closed connection" : 
			     strerror(io_errno));

	/*
	 * If we were never connected to the server,
	 * eg, they rejected us at registration time,
	 * then we will just move on to the next
	 * server in the list.
	 */
	if (sopen)
		get_connected(i, i);
	else
		get_connected(i + 1, i);

	from_server = primary_server;
	return 1;
}

int	server_data (int fd, int len, char *buffer)
{
	char *	end;
	Server *s;
	int	i = -1;

	while ((s = traverse_all_servers(&i)))
		if (s->des == fd)
			break;

	last_server = from_server = i;

	end = buffer + len;
	if (*--end == '\n')
		*end-- = 0;
	if (*end == '\r')
		*end-- = 0;

	if (x_debug & DEBUG_INBOUND)
		yell("[%d] <- [%s]", 
			SERVER(i)->des, buffer);

	parsing_server_index = i;
	parse_server(buffer);
	parsing_server_index = -1;
	reset_display_target();

	from_server = primary_server;
	return 0;
}

Discipline	server_discipline = {
	server_data,
	NULL,
	server_error,
	server_error
};


/* SERVER OUTPUT STUFF */
static void 	vsend_to_aserver (int, const char *format, va_list args);

void	send_to_aserver (int refnum, const char *format, ...)
{
	va_list args;

	va_start(args, format);
	vsend_to_aserver(refnum, format, args);
	va_end(args);
}

void	send_to_server (const char *format, ...)
{
	va_list args;
	va_start(args, format);
	vsend_to_aserver(from_server, format, args);
	va_end(args);
}

/* send_to_server: sends the given info the the server */
static void 	vsend_to_aserver (int server, const char *format, va_list args)
{
	char	buffer[BIG_BUFFER_SIZE * 3 + 1]; /* make this buffer *much*
						  * bigger than needed */
	size_t	size = BIG_BUFFER_SIZE * 3;
	Server *s;
	char	*buf = buffer;
	int	len,
		des;

	/* Obviously if they have nothing to say, then return */
	if (!format)
		return;			/* What-EVER */

	/* Default to primary_server if all else fails */
	if (server == -1)
		server = primary_server;

	/* 
	 * If primary_server is -1, then we are not connected to *any*
	 * server at all.  This is a serious condition and the user
	 * needs to be told.
	 */
	if (server == -1)
	{
                if (do_hook(DISCONNECT_LIST, "No Connection to %d", server))
			say("You are not connected to a server, "
				"use /SERVER to connect.");
		return;
	}

	/* Ok.  Grab the server entry */
	s = SERVER(server);

	/* If the server isn't open, then don't even bother. */
	if ((des = s->des) == -1)
		return;			/* What-EVER */

	/* Press the data to the output buffer */
	len = vsnprintf(buf, BIG_BUFFER_SIZE, format, args);

	/* Doovie it up any according to the user's wishes */
	if (outbound_line_mangler)
	{
		if (mangle_line(buffer, outbound_line_mangler, size) > size)
			yell("Mangle_line truncated results.  Ick!");
	}

	/* This is used for /redirect, don't remove this */
	s->sent = 1;

	/* 
	 * Crop the line back to RFC limitations, and make sure
	 * that it has the proper \r\n line termination.
	 */
	if (len > (IRCD_BUFFER_SIZE - 2) || len == -1)
		buffer[IRCD_BUFFER_SIZE - 2] = 0;
	if (x_debug & DEBUG_OUTBOUND)
		yell("[%d] -> [%s]", des, buffer);
	strmcat(buffer, "\r\n", IRCD_BUFFER_SIZE);

	block_server_delete++;
	if (do_hook(SEND_TO_SERVER_LIST, "%d %d %s", server, des, buffer))
	{
	    if (write(des, buffer, strlen(buffer)) == -1 && 
			(!get_int_var(NO_FAIL_DISCONNECT_VAR)))
	    {
		say("Write to server failed.  Closing connection.");
		close_server(server, strerror(errno));
	    }
	}
	block_server_delete--;
}


void	clear_sent_to_server (int i)
{
	CHECK_SERVER(i)
	SERVER(i)->sent = 0;
}

int	sent_to_server (int i)
{
	CHECK_SERVER_RET(i, -1)
	return SERVER(i)->sent;
}

void	flush_server (int i)
{
	CHECK_SERVER(i)
	if (!is_server_connected(i))
		return;
	set_server_redirect(i, "0");
	send_to_aserver(i, "%s", "***.");
}



/************************** CONNECTION STUFF ****************************/
static int 	connect_to_server_stage_2 (int fd, int len, char *buffer);
/*
 * This establishes a new connection to 'new_server'.  This function does
 * not worry about why or where it is doing this.  It is only concerned
 * with getting a connection up and running.
 *
 * NOTICE!  THIS MUST ONLY EVER BE CALLED BY connect_to_new_server()!
 * IF YOU CALL THIS ELSEWHERE, THINGS WILL BREAK AND ITS NOT MY FAULT!
 */
static int 	connect_to_server (int new_server)
{
	int 			des;
	struct sockaddr_in *	localaddr;
	struct sockaddr_in *	remaddr;
	int			len;
	u_short			this_sucks;
	Server *		s;

	/*
	 * Can't connect to refnum -1, this is definitely an error.
	 */
	if (new_server == -1)
	{
		say ("Connecting to refnum -1.  That makes no sense.");
		return -1;		/* XXXX */
	}

	CHECK_SERVER_RET(new_server, -1)
	s = get_server(new_server);

	/*
	 * If we are already connected to the new server, go with that.
	 */
	if (s->des != -1)
	{
		say("Connected to port %d of server %s", 
			s->port, get_server_name(new_server));
		from_server = new_server;
		return -2;		/* Server is already connected */
	}

	/*
	 * Make an attempt to connect to the new server.
	 */
	say("Connecting to port %d of server %s [refnum %d]", 
			s->port, get_server_name(new_server), new_server);

	s->closing = 0;
	oper_command = 0;
	errno = 0;

	this_sucks = htons((unsigned short)s->port);
	s->connection_in_progress = 1;
	des = connect_by_number(get_server_name(new_server), &this_sucks,
		SERVICE_CLIENT, PROTOCOL_TCP, connect_to_server_stage_2);
	if (des < 0)
	{
		if (x_debug & DEBUG_SERVER_CONNECT)
			say("new_des is %d", des);

		say("Unable to connect to port %d of server %s: %s", 
				s->port, get_server_name(new_server),
				my_strerror(errno));
		return -1;
	}

	s->port = ntohs(this_sucks);
	s->des = des;

	say("Nonblocking connect in progress...");
	block_server_delete++;
	while (s->connection_in_progress)	/* XXX Might be gone ??? */
		io("Waiting on connect");
	block_server_delete--;
	if (s->des == -1)		/* Oh boy.  It really died. */
	{
		say("Nonblocking connect failed.  Punting.");
		return -1;
	}

	say("Nonblocking connect on port %d of server %s complete",
			s->port, get_server_name(new_server));

	if (*get_server_name(new_server) != '/')
	{
	    my_bzero(s->local_sockname);
	    localaddr = &s->local_sockname;
	    my_bzero(s->remote_sockname);
	    remaddr = &s->remote_sockname;

	    len = sizeof(*remaddr);
	    if (getpeername(des, (struct sockaddr *)remaddr, &len))
	    {
		say("Attempt to connect failed :%s ", strerror(errno));
		if (x_debug & DEBUG_SERVER_CONNECT)
			say("new_des is %d", des);

		say("Unable to connect to port %d of server %s", 
			s->port, get_server_name(new_server));

		s->des = new_close(s->des);
		return -1;
	    }

	    len = sizeof(*localaddr);
	    getsockname(des, (struct sockaddr *)localaddr, &len);
	    s->local_addr = localaddr->sin_addr;

	    len = sizeof(*remaddr);
	    getpeername(des, (SA *)remaddr, &len);
	}

	from_server = new_server;			/* XXX sigh */
	line_buffering(des, &server_discipline);	/* XXX */

	/*
	 * Initialize all of the server_list data items
	 * XXX - Calling add_to_server_list is a hack.
	 */
	add_to_server_list(get_server_name(new_server), s->port, NULL, NULL, 1);
	s->oper = 0;
	if (!s->d_nickname)
		malloc_strcpy(&s->d_nickname, nickname);

	/*
	 * Reset everything and go on our way.
	 */
	reset_display_target();
	update_all_status();
	return 0;
}

static int 	connect_to_server_stage_2 (int fd, int len, char *buffer)
{
	int	i;

	for (i = 0; i < number_of_servers; i++)
	{
		if (SERVER(i)->des == fd)
		{
			SERVER(i)->connection_in_progress = 0;
			return 0;
		}
	}
	panic("connect_to_server_stage_2 called invalidly.");
}


/*
 * This attempts to substitute a new connection to server 'new_server'
 * for the server connection 'old_server'.  
 * If 'old_server' is -1, then this is the first connection for the entire
 * 	client.  We have to handle this differently because no windows exist
 *	at this point and we have to be careful not to call 
 *	window_check_servers.
 * If 'old_server' is -2, then this is a new server connection on an existing
 * 	window.  This is used when a window "splits off" from an existing
 *	server connection to a new server.
 */
int 	connect_to_new_server (int new_server, int old_server, int new_conn)
{
	int	x;
	int	old;

	/* Determine if this should be above or below the connect call */
	from_server = new_server;	/* XXX sigh */
	set_display_target(NULL, LOG_CRAP);

	/*
	 * First of all, if we can't connect to the new server, we don't
	 * do anything here.  Note that this might succeed because we are
	 * already connected to the new server, in which case 
	 * 'is_server_connected' should be true.
	 */
	if ((x = connect_to_server(new_server)) == -1)
	{
		if (old_server >= 0 && SERVER(old_server)->des != -1)
			say("Connection to server %s resumed...", 
				get_server_name(old_server));

		reset_display_target();
		return -1;
	}

	/*
	 * If connect_to_server() resulted in a new server connection,
	 * go ahead and register it.
	 */
	if (x == 0)
		register_server(new_server, SERVER(new_server)->d_nickname);

	/*
	 * If we're not actually switching servers...
	 */
	if (old_server == new_server)
	{
		/*
		 * If we were asked to connect to a server we were already
		 * connected to, then there's nothing more to be done, eh?
		 */
		if (x == -2)
			return 0;

		/*
	 	 * We must be reconnecting to a server.  Try to find and
		 * grab all of its old windows that are just lying around.
		 * This should be harmless at worst.
		 */
		change_window_server(old_server, new_server);
		return 0;
	}

	/*
	 * At this point, we know that new_server != old_server, so we are
	 * actually trying to change server, rather than just try to re-
	 * establish an old, dropped one.  We do some bookkeeping here to
	 * try to keep everything in order, as far as windows and channels go.
	 * This is not done for /window server, though.
	 */
	if (new_conn == 0)
	{
		if (is_server_open(old_server))
		{
			/* 
			 * We need to close off the old connection now.
			 */
			close_server(old_server, "changing servers");
		}

		/*
		 * We *always* move over any channels on the older server
		 * to the new server.
		 */
		channel_server_change(old_server, new_server);

		/*
		 * If 'new_conn' is 0, and 'old_server' is -1, then we're
		 * doing something like a /server in a disconnected window.
		 * That's no sweat.  We will just try to figure out what
		 * server this window was last connected to, and move all of
		 * those windows over to this new server.
		 */
		if (old_server == -1)
		{
			if (!(old = get_window_oldserver(0)) != -1)
				old_server = old;
		}

		/*
		 * In all situations, we always want to move all of the
		 * windows over from the last server to this new server 
		 * as a logical group.  Grab any old windows that seem 
		 * to want to be part of this fun as well.  Also, do a 
		 * window_check_servers() to make sure all is coherent.
		 */
		change_window_server(old_server, new_server);
	}

	/*
	 * And it never hurts, in any situation, to set the AWAY message
	 * if there is one.
	 */
	if (!get_server_away(new_server) && get_server_away(old_server))
		set_server_away(new_server, get_server_away(old_server));

	reset_display_target();
	return 0;
}


/*
 * 'get_connected' is a front end to 'connect_to_new_server' that repeatedly
 * tries as many servers as it takes to get a new server connection going.
 * If all else fails, it reverts back to the original connection, provided
 * that the original connection existed. ;-)
 */
int 	get_connected (int new, int old)
{
	int	i;
	int	j;
	int	connected;

	if (!SERVER(0))
	{
                if (do_hook(DISCONNECT_LIST, "No Server List"))
		    say("No server list. Use /SERVER to connect to a server");
		return -1;
	}

	connected = is_server_connected(old);

	/* Try all of the other servers, stop when one of them works. */
	for (i = 0; i < number_of_servers; i++)
	{
		j = (i + new) % number_of_servers;
		if (new != old && j == old)
			continue;
		if (connect_to_new_server(j, old, 0) == 0)
			return j;
	}

	/* If we reach this point, we have failed.  Time to punt */

	/*
	 * If our prior state was connected, revert back to the prior
	 * connected server.
	 */
	if (connected && new != old)
	{
		say("All attempts to establish a new server connection "
				"have failed.");
		say("Your previous server connection will be resumed.");
		from_server = old;
		return -1;
	}

	/*
	 * Otherwise, our prior state was unconnected.  Tell the user
	 * that we give up and tough luck.
	 */
	if (do_hook(DISCONNECT_LIST, "Unable to connect to a server"))
		say("Sorry, cannot connect.  "
		    "Use /SERVER to connect to a server");
	return -1;
}


int 	close_all_servers (const char *message)
{
	int i;

	for (i = 0; i < number_of_servers; i++)
		close_server(i, message);

	return 0;
}

/*
 * close_server: Given an index into the server list, this closes the
 * connection to the corresponding server.  If 'message' is anything other
 * than the NULL or the empty_string, it will send a protocol QUIT message
 * to the server before closing the connection.
 */
void	close_server (int old, const char *message)
{
	/* Make sure server refnum is valid */
	if (old < 0 || old >= number_of_servers)
	{
		yell("Closing server [%d] makes no sense!", old);
		return;
	}

	if (SERVER(old)->des == -1)
		return;			/* Can't close server that isn't open */
	if (SERVER(old)->closing == 1)
		return;			/* Close already pending */

	clean_server_queues(old);
	SERVER(old)->waiting_out = SERVER(old)->waiting_in = 0;

	/* Channels are *always* suspended */
	channel_server_disconnect(old);

	SERVER(old)->oper = 0;
	SERVER(old)->connection_in_progress = 0;	/* Just in case */
	SERVER(old)->registration_pending = 0;
	SERVER(old)->connected = 0;
	new_free(&SERVER(old)->nickname);
	new_free(&SERVER(old)->s_nickname);

	if (SERVER(old)->des != -1)
	{
		if (message && *message && !SERVER(old)->closing)
		{
		    SERVER(old)->closing = 1;
		    if (x_debug & DEBUG_OUTBOUND)
			yell("Closing server %d because [%s]", 
			   old, message ? message : empty_string);

		    send_to_aserver(old, "QUIT :%s\n", message);
		    do_hook(SERVER_LOST_LIST, "%d %s %s", 
				old, get_server_name(old), message);
		}

		SERVER(old)->des = new_close(SERVER(old)->des);
	}

	return;
}

/*
 * is_server_open: Returns true if the given server index represents a server
 * with a live connection, returns false otherwise 
 */
int	is_server_open (int i)
{
	CHECK_SERVER_RET(i, 0)
	if (SERVER(i)->connection_in_progress)
		return 0;	/* Not open yet */
	return (SERVER(i)->des != -1);
}

/*
 * is_server_connected: returns true if the given server is connected.  This
 * means that both the tcp connection is open and the user is properly
 * registered 
 */
int	is_server_connected (i)
{
	CHECK_SERVER_RET(i, 0)
	return SERVER(i)->connected;
}

/*
 * Informs the client that the user is now officially registered or not
 * registered on the specified server.
 */
void 	server_is_connected (int i, int value)
{
	CHECK_SERVER(i)

	SERVER(i)->connected = value;
	SERVER(i)->registration_pending = 0;
	if (value)
		SERVER(i)->eof = 0;
}


/********************* OTHER STUFF ************************************/

/* AWAY STATUS */
/*
 * Encapsulates everything we need to change our AWAY status.
 * This improves greatly on having everyone peek into that member.
 * Also, we can deal centrally with someone changing their AWAY
 * message for a server when we're not connected to that server
 * (when we do connect, then we send out the AWAY command.)
 * All this saves a lot of headaches and crashes.
 */
void	set_server_away (int i, const char *message)
{
	if (i == -1)
		say("You are not connected to a server.");

	else if (message && *message)
	{
		if (get_server_away(i) != message)
			malloc_strcpy(&get_server(i)->away, message);
		if (is_server_connected(i))
			send_to_aserver(i, "AWAY :%s", message);
	}
	else
	{
		new_free(&SERVER(i)->away);
		if (is_server_connected)
			send_to_aserver(i, "AWAY :");
	}
}

const char *	get_server_away (int i)
{
	if (i == -2)
	{
		int	i = -1;
		while (traverse_all_servers(&i))
		{
			if (is_server_connected && get_server_away(i))
				return get_server_away(i);
		}
		return NULL;
	}

	CHECK_SERVER_RET(i, NULL)
	return SERVER(i)->away;
}


/* USER MODES */
static char *	do_umode (int i)
{
	char *	c;
	long 	flags;
	long 	flags2;
	int 	j;

	CHECK_SERVER_RET(i, NULL)
 	c = SERVER(i)->umode;
	flags = SERVER(i)->flags;
	flags2 = SERVER(i)->flags2;

	for (j = 0; SERVER(i)->umodes[j]; j++)
	{
		if (j > 31)
		{
			if (flags2 & (0x1 << (j - 32)))
				*c++ = SERVER(i)->umodes[j];
		}
		else
		{
			if (flags & (0x1 << j))
				*c++ = SERVER(i)->umodes[j];
		}
	}

	*c = 0;
	return SERVER(i)->umode;
}

const char *	get_possible_umodes (int i)
{
	CHECK_SERVER_RET(i, empty_string)
	return SERVER(i)->umodes;
}

const char *	get_umode (int i)
{
	CHECK_SERVER_RET(i, empty_string)
	return SERVER(i)->umode;
}

void 	clear_user_modes (int i)
{
	CHECK_SERVER(i)
	SERVER(i)->flags = SERVER(i)->flags2 = 0;
	do_umode(i);
}

void	set_server_flag (int i, int flag, int value)
{
	CHECK_SERVER(i)

	if (flag > 31)
	{
		if (value)
			SERVER(i)->flags2 |= 0x1 << (flag - 32);
		else
			SERVER(i)->flags2 &= ~(0x1 << (flag - 32));
	}
	else
	{
		if (value)
			SERVER(i)->flags |= 0x1 << flag;
		else
			SERVER(i)->flags &= ~(0x1 << flag);
	}
	do_umode(i);
}

int	get_server_flag (int i, int value)
{
	CHECK_SERVER_RET(i, 0)
	if (value > 31)
		return SERVER(i)->flags2 & (0x1 << (value - 32));
	else
		return SERVER(i)->flags & (0x1 << value);
}

void	update_user_mode (Char *modes)
{
	int	onoff = 1;
	Char *	p_umodes = get_possible_umodes(from_server);

	if (x_debug & DEBUG_SERVER_CONNECT)
		yell("Possible user modes for server [%d]: [%s]", 
			from_server, p_umodes);

	for (; *modes; modes++)
	{
		if (*modes == '-')
			onoff = 0;
		else if (*modes == '+')
			onoff = 1;

		else if   ((*modes >= 'a' && *modes <= 'z')
			|| (*modes >= 'A' && *modes <= 'Z'))
		{
			size_t 	idx;
			int 	c = *modes;

			if (c == 'O')
				c = 'o';		/* XXXXX */

			idx = ccspan(p_umodes, c);
			if (p_umodes[idx] == 0)
				panic("Invalid user mode referenced");
			set_server_flag(from_server, idx, onoff);

			if (c == 'o')
				set_server_operator(from_server, onoff);
		}
	}
}

void	reinstate_user_modes (int serv)
{
	Char *modes = get_umode(serv);

	if (!modes && !*modes)
		modes = send_umode;

	if (modes && *modes)
	{
		if (x_debug & DEBUG_OUTBOUND)
			yell("Reinstating your user modes on server "
				"[%d] to [%s]", serv, modes);
		send_to_server("MODE %s +%s", get_server_nickname(serv), modes);
		clear_user_modes(serv);
	}
}

/* SERVER VERSIONS */
/*
 * set_server_version: Sets the server version for the given server type.  A
 * zero version means pre 2.6, a one version means 2.6 aso. (look server.h
 * for typedef)
 */
void	set_server_version (int i, int version)
{
	CHECK_SERVER(i)
	SERVER(i)->version = version;
}

/*
 * get_server_version: returns the server version value for the given server
 * index 
 */
int	get_server_version (int i)
{
	CHECK_SERVER_RET(i, 0)
	return SERVER(i)->version;
}

/* SERVER NAMES */
/* get_server_name: returns the name for the given server index */
Char *	get_server_name (int i)
{
	CHECK_SERVER_RET(i, empty_string)
	return SERVER(i)->name;
}

/* get_server_itsname: returns the server's idea of its name */
Char *	get_server_itsname (int i)
{
	CHECK_SERVER_RET(i, empty_string)

	if (SERVER(i)->itsname)
		return SERVER(i)->itsname;
	else
		return get_server_name(i);
}

void	set_server_itsname (int i, const char *name)
{
	CHECK_SERVER(i)
	malloc_strcpy(&SERVER(i)->itsname, name);
}

void	set_server_version_string (int i, const char *ver)
{
	CHECK_SERVER(i)
	malloc_strcpy(&SERVER(i)->version_string, ver);
}

Char *	get_server_version_string (int i)
{
	CHECK_SERVER_RET(i, empty_string)
	return SERVER(i)->version_string;
}


/* CONNECTION/REGISTRATION STATUS */
void	register_server (int i, const char *nickname)
{
	CHECK_SERVER(i)
	if (SERVER(i)->registration_pending)
		return;		/* Whatever */

	if (SERVER(i)->connected)
		return;		/* Whatever */

	SERVER(i)->registration_pending = 1;
	if (SERVER(i)->password)
		send_to_aserver(i, "PASS %s", SERVER(i)->password);

	send_to_aserver(i, "USER %s %s %s :%s", username, 
			(send_umode && *send_umode) ? send_umode : 
			(LocalHostName ? LocalHostName : hostname), 
			username, (*realname ? realname : space));
	change_server_nickname(i, nickname);
}

/*
 * password_sendline: called by send_line() in get_password() to handle
 * hitting of the return key, etc 
 * -- Callback function
 */
void 	password_sendline (char *data, char *line)
{
	int	new_server;

	if (line && *line)
	{
		new_server = atoi(next_arg(line, &line));
		set_server_password(new_server, line);
		connect_to_new_server(new_server, new_server, 1);
	}
}

/*
 * set_server_password: this sets the password for the server with the given
 * index.  If password is null, the password for the given server is returned 
 */
Char *	set_server_password (int i, Char *password)
{
	CHECK_SERVER_RET(i, NULL)
	if (password)
		malloc_strcpy(&SERVER(i)->password, password);
	return (SERVER(i)->password);
}


int 	auto_reconnect_callback (void *d)
{
	char *	stuff = (char *) d;
	int	servref;

	servref = my_atol(stuff);
	new_free((char **)&d);

	get_connected(servref, servref);
	return 0;
}

/* PORTS */
/* get_server_port: Returns the connection port for the given server index */
int	get_server_port (int i)
{
	CHECK_SERVER_RET(i, 0);
	return get_server(i)->port;
}

int	get_server_local_port (int i)
{
	CHECK_SERVER_RET(i, 0);
	return ntohs(get_server(i)->local_sockname.sin_port);
}

struct in_addr	get_server_local_addr (int i)
{
	struct in_addr 	y;
	CHECK_SERVER_RET(i, y);
	return get_server(i)->local_addr;
}

struct in_addr	get_server_uh_addr (int i)
{
	struct in_addr 	y;
	CHECK_SERVER_RET(i, y);
	return get_server(i)->uh_addr;
}

/* USERHOST */
/*
 * get_server_userhost: return the userhost for this connection to server
 */
const char *	get_server_userhost (int i)
{
	CHECK_SERVER_RET(i, empty_string);
	return get_server(i)->userhost;
}

/*
 * got_my_userhost -- callback function, XXXX doesnt belong here
 */
void 	got_my_userhost (UserhostItem *item, char *nick, char *stuff)
{
	Server	*s = SERVER(from_server);

	new_free(&s->userhost);
	s->userhost = m_3dup(item->user, "@", item->host);
	s->uh_addr = inet_anyton(item->host);
	if (ADDRESS(s->uh_addr) == INADDR_NONE)
	{
		yell("Ack.  The server says your userhost is [%s] and "
		     "I can't figure out the IP address of that host! "
		     "You won't be able to use /SET DCC_USE_GATEWAY_ADDR ON "
		     "with this server connection!", item->host);
	}
}



/* SERVER OPERATOR */
/*
 * get_server_operator: returns true if the user has op privs on the server,
 * false otherwise 
 */
int	get_server_operator (int i)
{
	CHECK_SERVER_RET(i, 0);
	return get_server(i)->oper;
}

/*
 * set_server_operator: If flag is non-zero, marks the user as having op
 * privs on the given server.  
 */
void	set_server_operator (int i, int flag)
{
	CHECK_SERVER(i)
	get_server(i)->oper = flag;
	oper_command = 0;		/* No longer doing oper */
	do_umode(i);
}



/* COOKIES */
void	set_server_cookie (int i, const char *cookie)
{
	CHECK_SERVER(i)
	if (get_server(i)->cookie)
		send_to_aserver(i, "COOKIE %s", get_server(i)->cookie);
	malloc_strcpy(&get_server(i)->cookie, cookie);
}

/* MOTD */
void 	set_server_motd (int i, int flag)
{
	CHECK_SERVER(i)
	get_server(i)->motd = flag;
}

int 	get_server_motd (int i)
{
	CHECK_SERVER_RET(i, 0)
	return get_server(i)->motd;
}



/* NICKNAMES */
/*
 * get_server_nickname: returns the current nickname for the server index 
 */
const char *	get_server_nickname (int i)
{
	if (i == -1)
		return "<not registered yet>";

	CHECK_SERVER_RET(i, empty_string)
	return get_server(i)->nickname;
}

const char *	get_real_server_nickname (int i)
{
	CHECK_SERVER_RET(i, NULL)
	return get_server(i)->nickname;
}

int	is_me (int i, const char *nick)
{
	if (i == -1 && from_server != -1)
		i = from_server;

	CHECK_SERVER_RET(i, 0)

	if (nick && get_real_server_nickname(i))
		return !my_stricmp(nick, get_real_server_nickname(i));

	return 0;
}

/*
 * This is the function to attempt to make a nickname change.  You
 * cannot send the NICK command directly to the server: you must call
 * this function.  This function makes sure that the neccesary variables
 * are set so that if the NICK command fails, a sane action can be taken.
 *
 * If ``nick'' is NULL, then this function just tells the server what
 * we're trying to change our nickname to.  If we're not trying to change
 * our nickname, then this function does nothing.
 */
void	change_server_nickname (int i, const char *nick)
{
	Server *s;
	char *	n;

	CHECK_SERVER(i)
	s = SERVER(i);

	s->resetting_nickname = 0;
	if (nick)
	{
		n = LOCAL_COPY(nick);
		if ((n = check_nickname(n, 1)) != NULL)
		{
			malloc_strcpy(&s->d_nickname, n);
			malloc_strcpy(&s->s_nickname, n);
		}
		else
			reset_nickname(i);
	}

	if (s->s_nickname)
		send_to_aserver(i, "NICK %s", s->s_nickname);
}

const char *	get_pending_nickname (int i)
{
	CHECK_SERVER_RET(i, NULL)
	return get_server(i)->s_nickname;
}

void	accept_server_nickname (int i, const char *nick)
{
	CHECK_SERVER(i)
	malloc_strcpy(&get_server(i)->nickname, nick);
	malloc_strcpy(&get_server(i)->d_nickname, nick);
	new_free(&get_server(i)->s_nickname);
	get_server(i)->fudge_factor = 0;

	if (i == primary_server)
		strmcpy(nickname, nick, NICKNAME_LEN);

	update_all_status();
}

void	nick_command_is_pending (int i, int value)
{
	CHECK_SERVER(i)
	get_server(i)->nickname_pending = value;
}

int	is_nick_command_pending (int i)
{
	CHECK_SERVER_RET(i, 0)
	return get_server(i)->nickname_pending;
}

/* 
 * This will generate up to 18 nicknames plus the 9-length(nickname)
 * that are unique but still have some semblance of the original.
 * This is intended to allow the user to get signed back on to
 * irc after a nick collision without their having to manually
 * type a new nick every time..
 * 
 * The func will try to make an intelligent guess as to when it is
 * out of guesses, and if it ever gets to that point, it will do the
 * manually-ask-you-for-a-new-nickname thing.
 */
void 	fudge_nickname (int servnum)
{
	char 	l_nickname[NICKNAME_LEN + 1];
	Server *s = SERVER(servnum);

	if (!s)
		return;		/* Don't bother. */

	/*
	 * If we got here because the user did a /NICK command, and
	 * the nick they chose doesnt exist, then we just dont do anything,
	 * we just cancel the pending action and give up.
	 */
	if (s->nickname_pending)
	{
		new_free(&s->s_nickname);
		return;
	}

	/*
	 * Ok.  So we're not doing a /NICK command, so we need to see
	 * if maybe we're doing some other type of NICK change.
	 */
	if (s->s_nickname)
		strlcpy(l_nickname, s->s_nickname, NICKNAME_LEN);
	else if (s->nickname)
		strlcpy(l_nickname, s->nickname, NICKNAME_LEN);
	else
		strlcpy(l_nickname, nickname, NICKNAME_LEN);


	if (s->fudge_factor < strlen(l_nickname))
		s->fudge_factor = strlen(l_nickname);
	else
	{
		if (++s->fudge_factor == 17)
		{
			/* give up... */
			reset_nickname(servnum);
			s->fudge_factor = 0;
			return;
		}
	}

	/* 
	 * Process of fudging a nickname:
	 * If the nickname length is less then 9, add an underscore.
	 */
	if (strlen(l_nickname) < 9)
		strcat(l_nickname, "_");

	/* 
	 * The nickname is 9 characters long. roll the nickname
	 */
	else
	{
		char tmp = l_nickname[8];
		l_nickname[8] = l_nickname[7]; l_nickname[7] = l_nickname[6];
		l_nickname[6] = l_nickname[5]; l_nickname[5] = l_nickname[4];
		l_nickname[4] = l_nickname[3]; l_nickname[3] = l_nickname[2];
		l_nickname[2] = l_nickname[1]; l_nickname[1] = l_nickname[0];
		l_nickname[0] = tmp;
	}

	/*
	 * This is the degenerate case
	 */
	if (!strcmp(l_nickname, "_________"))
	{
		reset_nickname(servnum);
		return;
	}

	change_server_nickname(servnum, l_nickname);
}


/*
 * -- Callback function
 */
void 	nickname_sendline (char *data, char *nick)
{
	int	new_server;

	new_server = atoi(data);
	change_server_nickname(new_server, nick);
}

/*
 * reset_nickname: when the server reports that the selected nickname is not
 * a good one, it gets reset here. 
 * -- Called by more than one place
 */
void 	reset_nickname (int i)
{
	char	server_num[10];

	CHECK_SERVER(i)
	if (SERVER(i)->resetting_nickname == 1)
		return;		/* Don't repeat the reset */

	SERVER(i)->resetting_nickname = 1;
	say("You have specified an invalid nickname");
	if (!dumb_mode)
	{
		say("Please enter your nickname");
		strcpy(server_num, ltoa(i));
		add_wait_prompt("Nickname: ", nickname_sendline, server_num,
			WAIT_PROMPT_LINE, 1);
	}
	update_all_status();
}


/* REDIRECT STUFF */
void 	set_server_redirect (int s, const char *who)
{
	malloc_strcpy(&get_server(s)->redirect, who);
}

const char *get_server_redirect (int s)
{
	return get_server(s)->redirect;
}

int	check_server_redirect (const char *who)
{
	if (!who || !FROMSERV->redirect)
		return 0;

	if (!strncmp(who, "***", 3) && 
	    !strcmp(who + 3, FROMSERV->redirect))
	{
		set_server_redirect(from_server, NULL);
		if (!strcmp(who + 3, "0"))
			say("Server flush done.");
		return 1;
	}

	return 0;
}

/* WAIT STUFF */
/*
 * This isnt a command, its used by the wait command.  Since its extern,
 * and it doesnt use anything static in this file, im sure it doesnt
 * belong here.
 */
void 	server_hard_wait (int i)
{
	CHECK_SERVER(i)

	if (!is_server_connected(i))
		return;

	get_server(i)->waiting_out++;
	lock_stack_frame();
	send_to_aserver(i, "%s", lame_wait_nick);
	while (get_server(i) && (get_server(i)->waiting_in < 
				 get_server(i)->waiting_out))
		io("oh_my_wait");
}

void	server_passive_wait (const char *stuff)
{
	WaitCmd	*new_wait;
	int	i = from_server;

	new_wait = (WaitCmd *)new_malloc(sizeof(WaitCmd));
	new_wait->stuff = m_strdup(stuff);
	new_wait->next = NULL;

	if (get_server(i)->end_wait_list)
		get_server(i)->end_wait_list->next = new_wait;
	get_server(i)->end_wait_list = new_wait;
	if (!get_server(i)->start_wait_list)
		get_server(i)->start_wait_list = new_wait;

	send_to_aserver(i, "%s", wait_nick);
}

/*
 * How does this work?  Well, when we issue the /wait command it increments
 * a variable "waiting_out" which is the number of times that wait has been
 * caled so far.  If we get a wait token, we increase the waiting_in level
 * by one, and if the number of inbound waiting tokens is the same as the 
 * number of outbound tokens, then we are free to clear this stack frame
 * which will cause all of the pending waits to just fall out.
 */
int	server_check_wait (const char *nick)
{
	Server	*s = FROMSERV;

	if ((s->waiting_out > s->waiting_in) && !strcmp(nick, lame_wait_nick))
	{
		s->waiting_in++;
		unlock_stack_frame();
	        return 1;
	}

	if (s->start_wait_list && !strcmp(nick, wait_nick))
	{
		WaitCmd *old = s->start_wait_list;

		s->start_wait_list = old->next;
		if (old->stuff)
		{
			parse_line("WAIT", old->stuff, empty_string, 0, 0);
			new_free(&old->stuff);
		}
		if (s->end_wait_list == old)
			s->end_wait_list = NULL;
		new_free((char **)&old);
		return 1;
	}
	return 0;
}

int	doing_privmsg (void)
{
	return FROMSERV->doing_privmsg;
}

int	doing_notice (void)
{
	return FROMSERV->doing_notice;
}

int	doing_ctcp (void)
{
	return FROMSERV->in_ctcp_flag;
}

void	set_doing_privmsg (int v)
{
	FROMSERV->doing_privmsg = v;
}

void	set_doing_notice (int v)
{
	FROMSERV->doing_notice = v;
}

void	set_doing_ctcp (int v)
{
	FROMSERV->in_ctcp_flag = v;
}

/*****/
void	set_server_invite_channel (Char *chan)
{
	malloc_strcpy(&FROMSERV->invite_channel, chan);
}

Char *	get_server_invite_channel (void)
{
	return FROMSERV->invite_channel;
}

void	set_server_last_notify (Char *nick)
{
	malloc_strcpy(&FROMSERV->last_notify_nick, nick);
}

Char *	get_server_last_notify (void)
{
	return FROMSERV->last_notify_nick;
}

void	set_server_joined_nick (Char *nick)
{
	malloc_strcpy(&FROMSERV->joined_nick, nick);
}

Char *	get_server_joined_nick (void)
{
	return FROMSERV->joined_nick;
}

void	set_server_public_nick (Char *nick)
{
	malloc_strcpy(&FROMSERV->public_nick, nick);
}

Char *	get_server_public_nick (void)
{
	return FROMSERV->public_nick;
}

void	set_server_recv_nick (Char *nick)
{
	malloc_strcpy(&FROMSERV->recv_nick, nick);
}

Char *	get_server_recv_nick (void)
{
	return FROMSERV->recv_nick;
}

void	set_server_sent_nick (Char *nick)
{
	malloc_strcpy(&FROMSERV->sent_nick, nick);
}

Char *	get_server_sent_nick (void)
{
	return FROMSERV->sent_nick;
}

void	set_server_sent_body (Char *text)
{
	malloc_strcpy(&FROMSERV->sent_body, text);
}

Char *	get_server_sent_body (void)
{
	return FROMSERV->sent_body;
}

struct ChannelListStru *	get_server_channels (Servref server)
{
	CHECK_SERVER_RET(server, NULL)
	return get_server(server)->channels;
}

/*
 * save_servers; dumps your serverlist to a file, which can be /load'ed
 * later.  --SrfRoG
 */
void  save_servers (FILE *fp)
{
	int     i;

	if (!server_list)
		return;         /* no servers */

	for (i = 0; i < number_of_servers; i++)
	{
		/* SERVER -ADD server:port:password:nick */
		fprintf(fp, "SERVER -ADD %s:%d:%s:%s\n",
			get_server_name(i),
			get_server_port(i),
			SERVER(i)->password,
			get_real_server_nickname(i) ? 
				get_real_server_nickname(i) : empty_string);
	}
}

void	set_server_window_count (Servref s, int count)
{
	get_server(s)->windows = count;
}

